﻿using Connected.Extensions;
using Connected.Utilities;
using Microsoft.AspNetCore.Components;
using Microsoft.AspNetCore.Components.Web;
using System.Timers;

namespace Connected.Components;

public partial class Input<T> : InputBase<T>
{

	#region Event Callbacks

	/// <summary>
	/// Paste hook for descendants.
	/// </summary>
#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously

	protected virtual async Task OnPaste(ClipboardEventArgs args)
#pragma warning restore CS1998 // Async method lacks 'await' operators and will run synchronously
	{
		// do nothing
		return;
	}

	/// <summary>
	/// callback to be called when the debounce interval has elapsed
	/// receives the Text as a parameter
	/// </summary>
	[Parameter] public EventCallback<string> OnDebounceIntervalElapsed { get; set; }

	protected Task OnDebounceChange()
	{

		if (TextChangeDelay > 0 && _timer != null)
		{
			_timer.Stop();
			return base.UpdateValuePropertyAsync(false);
		}
		return Task.CompletedTask;
	}

	protected override Task UpdateValuePropertyAsync(bool updateText)
	{
		// This method is called when Value property needs to be refreshed from the current Text property, so typically because Text property has changed.
		// We want to debounce only text-input, not a value being set, so the debouncing is only done when updateText==false (because that indicates the
		// change came from a Text setter)
		
		if (updateText)
		{
			// we have a change coming not from the Text setter, no debouncing is needed
			return base.UpdateValuePropertyAsync(updateText);
		}
		// if debounce interval is 0 we update immediately
		if (TextChangeDelay <= 0 || _timer == null)
			return base.UpdateValuePropertyAsync(updateText);
		// If a debounce interval is defined, we want to delay the update of Value property.
		_timer.Stop();
		// restart the timer while user is typing
		_timer.Start();
		return Task.CompletedTask;
	}

	private void OnTimerTick(object sender, ElapsedEventArgs e)
	{
		InvokeAsync(OnTimerTickGuiThread).AndForget();
	}

	private async Task OnTimerTickGuiThread()
	{
		await base.UpdateValuePropertyAsync(false);
		await OnDebounceIntervalElapsed.InvokeAsync(Text);
	}
	protected Task OnInput(ChangeEventArgs args)
	{
		var input = args.Value.ToString();
		/*if (InputType==InputType.Number)
		{
			if (!Helper.IsNumeric(input))
			{
				input = Regex.Replace(input, "[^0-9.]", "");
				Text = input;
				UpdateValuePropertyAsync(true);
			}
		}*/
		if (!ChangeTextImmediately)
			return Task.CompletedTask;
		_isFocused = true;
		return SetTextAsync(input);
	}

	protected virtual void OnKeyDown(KeyboardEventArgs obj)
	{
		if (InputType == InputType.Number)
		{
			if (Helper.IsNumeric(obj.Key))
			{
				_isFocused = true;
				base.OnKeyDown.InvokeAsync(obj).AndForget();
			}
		} else
		{
			_isFocused = true;
			base.OnKeyDown.InvokeAsync(obj).AndForget();
		}
	}

	protected async Task OnChange(ChangeEventArgs args)
	{
		if (TextChangeDelay > 0 && _timer != null)
		{
			_timer.Stop();
			await UpdateValuePropertyAsync(false);
		}
		else
		{
			_internalText = args?.Value as string;
			await OnInternalInputChanged.InvokeAsync(args);
			if (!ChangeTextImmediately)
			{
				await SetTextAsync(args?.Value as string);
			}
		}
	}

	/// <summary>
	/// Invokes the callback when the Up arrow button is clicked when the input is set to <see cref="InputType.Number"/>.
	/// Note: use the optimized control <see cref="NumericField{T}"/> if you need to deal with numbers.
	/// </summary>
	[Parameter] public EventCallback OnIncrement { get; set; }

	/// <summary>
	/// Invokes the callback when the Down arrow button is clicked when the input is set to <see cref="InputType.Number"/>.
	/// Note: use the optimized control <see cref="NumericField{T}"/> if you need to deal with numbers.
	/// </summary>
	[Parameter] public EventCallback OnDecrement { get; set; }

	/// <summary>
	/// Button click event for clear button. Called after text and value has been cleared.
	/// </summary>
	[Parameter] public EventCallback<MouseEventArgs> OnClearButtonClick { get; set; }

	/// <summary>
	/// Mouse wheel event for input.
	/// </summary>
	[Parameter] public EventCallback<WheelEventArgs> OnMouseWheel { get; set; }

	public override async ValueTask FocusAsync()
	{
		try
		{
			if (InputType == InputType.Hidden && ChildContent != null)
				await _elementReference1.FocusAsync();
			else
				await ElementReference.FocusAsync();
		}
		catch (Exception e)
		{
			Console.WriteLine("Input.FocusAsync: " + e.Message);
		}
	}

	public override ValueTask BlurAsync()
	{
		return ElementReference.BlurAsync();
	}

	public override ValueTask SelectAsync()
	{
		return ElementReference.SelectAsync();
	}

	public override ValueTask SelectRangeAsync(int pos1, int pos2)
	{
		return ElementReference.SelectRangeAsync(pos1, pos2);
	}

	private void UpdateClearable(object value)
	{
		var showClearable = Clearable && ((value is string stringValue && !string.IsNullOrWhiteSpace(stringValue)) || (value is not string && value is not null));
		if (_showClearable != showClearable)
			_showClearable = showClearable;
	}

	protected override async Task UpdateTextPropertyAsync(bool updateValue)
	{
		await base.UpdateTextPropertyAsync(updateValue);
		if (Clearable)
			UpdateClearable(Text);
	}

	protected virtual async Task ClearButtonClickHandlerAsync(MouseEventArgs e)
	{
		await SetTextAsync(string.Empty, updateValue: true);
		await ElementReference.FocusAsync();
		await OnClearButtonClick.InvokeAsync(e);
	}

	/// <summary>
	/// Custom input Attributes
	/// </summary>
	[Parameter] public Dictionary<string, object>? Attributes { get; set; } = null;

	#endregion

	#region Style properties

	/// <summary>
	/// Type of the input element. It should be a valid HTML5 input type.
	/// </summary>
	[Parameter] public InputType InputType { get; set; } = InputType.Text;

	private string ClearIcon = Icons.Material.Filled.Clear;

	private string NumericUpIcon = Icons.Material.Filled.KeyboardArrowUp;

	private string NumericDownIcon = Icons.Material.Filled.KeyboardArrowDown;

	/// <summary>
	/// Hides the spin buttons"/>
	/// </summary>
	[Parameter] public bool HideSpinButtons { get; set; } = true;

	internal override InputType GetInputType()
	{
		return InputType;
	}

	/// <summary>
	/// Show clear button.
	/// </summary>
	[Parameter] public bool Clearable { get; set; } = false;

	#region Wrapper class
	/*[Parameter]
	public string WrapperClass { get; set; } = string.Empty;*/

	/*protected string WrapperClass => InputCssHelper.GetClassname(this,
				() => HasNativeHtmlPlaceholder() || !string.IsNullOrEmpty(Text) || Adornment == Adornment.Start || !string.IsNullOrWhiteSpace(Placeholder));*/

	[Parameter]
	public string ClassList { get; set; } = string.Empty;

	protected string WrapperClassList
	{
		get
		{
			/*return new CssBuilder("input")
		  .AddClass($"input-{Variant.ToDescription()}")
		  .AddClass($"input-adorned-{Adornment.ToDescription()}", Adornment != Adornment.None)
		  .AddClass($"input-margin-{Margin.ToDescription()}", when: () => Margin != Margin.None)
		  .AddClass($"input-underline", when: () => DisableUnderLine == false && Variant != Variant.Outlined)
		  .AddClass($"shrink", when: HasNativeHtmlPlaceholder() || !string.IsNullOrEmpty(Text) || Adornment == Adornment.Start || !string.IsNullOrWhiteSpace(Placeholder))
		  .AddClass($"disabled", Disabled)
		  .AddClass($"input-error", HasErrors)
		  .AddClass($"ltr", GetInputType() == InputType.Email || GetInputType() == InputType.Telephone)
		  .AddClass(WrapperClass)
		  .Build();*/

			var clasList =  new CssBuilder("input")
			.AddClass($"input-{Variant.ToDescription()}")
			.AddClass($"input-adorned-{Adornment.ToDescription()}", Adornment != Adornment.None)
			.AddClass($"input-margin-{Margin.ToDescription()}", when: () => Margin != Margin.None)
			.AddClass("input-underline", when: () => DisableUnderLine == false && Variant != Variant.Outlined)
			//.AddClass("shrink", when: () => HasNativeHtmlPlaceholder() || !string.IsNullOrEmpty(Text) || Adornment == Adornment.Start || !string.IsNullOrWhiteSpace(Placeholder))
			.AddClass("disabled", Disabled)
			.AddClass("input-error", HasErrors)
			.AddClass("ltr", GetInputType() == InputType.Email || GetInputType() == InputType.Telephone)
			.AddClass(ClassList)
			.Build();
			return clasList;

		}
	}

	/// <summary>
	/// User styles, applied on top of the component's own classes and styles.
	/// </summary>
	[Parameter]
	public string Style { get; set; }


	#endregion

	#region Input field class

	[Parameter]
	public string InputClass { get; set; } = string.Empty;
	protected string InputClassList
	{
		get
		{
			return new CssBuilder("input-slot")
			.AddClass("input-root")
			.AddClass($"input-root-{Variant.ToDescription()}")
			.AddClass($"input-root-adorned-{Adornment.ToDescription()}", Adornment != Adornment.None)
			.AddClass($"input-root-margin-{Margin.ToDescription()}", when: () => Margin != Margin.None)
			.AddClass(InputClass)
			.Build();

			//return new CssBuilder().AddClass(InputCssHelper.GetInputClassname(this)).Build();
		}
	}
	#endregion

	#region Adornment class
	[Parameter]
	public string AdornmentClass { get; set; } = string.Empty;
	protected string AdornmentClassList
	{
		get
		{
			return new CssBuilder("input-adornment")
			.AddClass($"input-adornment-{Adornment.ToDescription()}", Adornment != Adornment.None)
			.AddClass($"text", !string.IsNullOrEmpty(AdornmentText))
			.AddClass($"input-root-filled-shrink", Variant == Variant.Filled)
			.AddClass(AdornmentClass)
			.Build();

			//return new CssBuilder().AddClass(InputCssHelper.GetAdornmentClassname(this)).Build();
		}
	}
	#endregion

	#region Clear icon class
	[Parameter]
	public string ClearButtonClass { get; set; } = string.Empty;
	protected string ClearButtonClassList
	{
		get
		{
			return new CssBuilder()
				.AddClass("me-n1", Adornment == Adornment.End && HideSpinButtons == false)
				.AddClass("icon-button-edge-end", Adornment == Adornment.End && HideSpinButtons == true)
				.AddClass("me-6", Adornment != Adornment.End && HideSpinButtons == false)
				.AddClass("micon-button-edge-margin-end", Adornment != Adornment.End && HideSpinButtons == true)
				.AddClass(ClearButtonClass)
				.Build();
		}
	}
	#endregion

	#region Helper container class

	/// <summary>
	/// A space separated list of class names, added on top of the default helper container class list.
	/// </summary>
	[Parameter]
	public string? HelperContainerClass { get; set; }
	protected string HelperContainerClassList
	{
		get
		{
			return new CssBuilder("input-control-helper-container")
			.AddClass($"px-1", Variant == Variant.Filled)
			.AddClass($"px-2", Variant == Variant.Outlined)
			.AddClass($"px-1", Variant == Variant.Text)
			.AddClass(HelperContainerClass)
			.Build();
		}
	}

	#endregion

	#region Error container class -- still needs implementation

	#endregion

	#region Counter container class -- still needs implementation

	#endregion

	#endregion

	#region Content placeholders

	/// <summary>
	/// The current character counter, displayed below the text field.
	/// </summary>
	[Parameter] public bool ShowCharacterCounter { get; set; }

	/// <summary>
	/// The current character counter, displayed below the text field.
	/// </summary>
	public string GetCounterText()
	{
		if (ShowCharacterCounter)
			return Text.Length.ToString();
		return string.Empty;
	}

	private System.Timers.Timer _timer;

	private double _textChangeInterval;

	[Parameter]
	public int Lines { get; set; } = 1;


	/// <summary>
	/// Interval to be awaited in MILLISECONDS before changing the Text value
	/// </summary>
	[Parameter]
	public double TextChangeDelay
	{
		get => _textChangeInterval;
		set
		{
			if (NumericConverter<double>.AreEqual(_textChangeInterval, value))
				return;
			_textChangeInterval = value;
			if (_textChangeInterval == 0)
			{
				// not debounced, dispose timer if any
				ClearTimer(suppressTick: false);
				return;
			}
			SetTimer();
		}
	}

	private void SetTimer()
	{
		if (_timer == null)
		{
			_timer = new System.Timers.Timer();
			_timer.Elapsed += OnTimerTick;
			_timer.AutoReset = false;
		}
		_timer.Interval = TextChangeDelay;
	}

	private void ClearTimer(bool suppressTick = false)
	{
		if (_timer == null)
			return;
		var wasEnabled = _timer.Enabled;
		_timer.Stop();
		_timer.Elapsed -= OnTimerTick;
		_timer.Dispose();
		_timer = null;
		if (wasEnabled && !suppressTick)
			OnTimerTickGuiThread().AndForget();
	}

	/// <summary>
	/// ChildContent of the MudInput will only be displayed if InputType.Hidden and if its not null.
	/// </summary>
	[Parameter] public RenderFragment ChildContent { get; set; }

	private string TextCounter()
	{
		if (string.IsNullOrEmpty(Text)) return "0";
		return Text.Length.ToString();
	}

	[Parameter]
	public double Step { get; set; } = 1;
	//internal override InputType GetInputType() => InputType;
	protected string InputTypeString => InputType.ToDescription();
	public ElementReference ElementReference { get; private set; }
	private ElementReference _elementReference1;
	private Size GetButtonSize() => Margin == Margin.Dense ? Size.Small : Size.Medium;
	private bool _showClearable;
	private string _internalText;

	/// <summary>
	/// Sets the input text from outside programmatically
	/// </summary>
	/// <param name="text"></param>
	/// <returns></returns>
	public Task SetText(string text)
	{
		_internalText = text;
		return SetTextAsync(text);
	}

	private bool HasNativeHtmlPlaceholder()
	{
		return GetInputType() is InputType.Color or InputType.Date or InputType.DateTimeLocal or InputType.Month
			 or InputType.Time or InputType.Week;
	}

	#endregion

	#region Lifecycle events
	public override async Task SetParametersAsync(ParameterView parameters)
	{
		await base.SetParametersAsync(parameters);
		//if (!_isFocused || _forceTextUpdate)
		//    _internalText = Text;
		if (RuntimeLocation.IsServerSide && TextUpdateSuppression)
		{
			// Text update suppression, only in BSS (not in WASM).
			// This is a fix for #1012
			if (!_isFocused || _forceTextUpdate)
				_internalText = Text;
		}
		else
		{
			// in WASM (or in BSS with TextUpdateSuppression==false) we always update
			_internalText = Text;
		}

		if (Helper.IsNumericType(typeof(T).Name) && InputType != InputType.Number)
		{
			InputType = InputType.Number;
		}

	}

	protected override void OnParametersSet()
	{
		base.OnParametersSet();
		// if input is to be debounced, makes sense to bind the change of the text to oninput
		// so we set ChangeTextImmediately to true
		if (TextChangeDelay > 0)
			ChangeTextImmediately = true;
	}
	#endregion

}
